#include "rei-json/Array.h"
#include "rei-json/Field.h"
#include "rei-json/json.h"
#include <algorithm>
#include <cctype>
#include <charconv>
#include <chrono>
#include <memory>
#include <ostream>
#include <print>
#include <stdexcept>
#include <string>
#include <string_view>
#include <variant>
namespace  rei::json {

	void goToNextDoubleQuote(const std::string& str, int* const i) {
		while (str[*i] != '"') {
			if (str[*i] == '\\') {
				++(*i);
			}
			++(*i);
		}
	}

	void skipDigits(const std::string& str, int* const i) {
		while (isdigit(str[*i]) || str[*i] == '-') {
			++(*i);
		}
	}

	void skipWhiteSpaces(const std::string& str, int* const i) {
		while (str[*i] == ' ' || str[*i] == '\t' || str[*i] == '\n' || str[*i] == '\v') {
			++(*i);
		}
	}

	void parseObject(const std::string &str, int *i, JsonObject* object);
	void parseArray(const std::string &str, int *i, JsonArray* object);

	void parseObjectField(const std::string &str, int *i, JsonObject* object) {
		if (str[*i] != '"') {
			throw "wrong json: expected opening quote";
		}
		++(*i);
		const int keyStart = *i;
		goToNextDoubleQuote(str, i);
		const int keyLength = *i - keyStart;
		const std::string key = str.substr(keyStart, keyLength);
		++(*i);
		skipWhiteSpaces(str, i);
		if (str[*i] != ':') {
			throw "wrong json: expect 2 dots";
		}
		++(*i);
		skipWhiteSpaces(str, i);

		if (str[*i] == '"') {
			++(*i);
			const int valueStart = *i;
			goToNextDoubleQuote(str, i);
			const int valueLength = *i - valueStart;
			const std::string valueStr = str.substr(valueStart, valueLength);
			object->set(key, std::move(valueStr));
			++(*i);
		} else if (isdigit(str[*i]) || str[*i] == '-') {
			const int valueStart = *i;
			skipDigits(str, i);
			const int valueLength = *i - valueStart;
			const std::string valueStr = str.substr(valueStart, valueLength);
			int value;
			std::from_chars(str.c_str() + valueStart, str.c_str() + valueStart + valueLength, value);
			object->set(key, value);
		} else if (std::string_view{str}.substr(*i, 4) == "true") {
			const int valueStart = *i;
			*i += 4;
			const int valueLength = *i - valueStart;
			object->set(key, true);
		} else if (std::string_view{str}.substr(*i, 5) == "false") {
			const int valueStart = *i;
			*i += 5;
			const int valueLength = *i - valueStart;
			object->set(key, false);
		} else if (std::string_view{str}.substr(*i, 4) == "null") {
			*i += 4;
			object->setNull(key);
		} else if (str[*i] == '{'){
			JsonObject newObject{}; 
			parseObject(str, i, &newObject);
			object->set(key, std::move(newObject));
		} else if (str[*i] == '['){
			JsonArray newArray{}; 
			parseArray(str, i, &newArray);
			object->set(key, std::move(newArray));
		}
	}

	void parseArrayField(const std::string &str, int *i, JsonArray* array) {
		skipWhiteSpaces(str, i);
		if (str[*i] == '"') {
			++(*i);
			const int valueStart = *i;
			goToNextDoubleQuote(str, i);
			const int valueLength = *i - valueStart;
			const std::string valueStr = str.substr(valueStart, valueLength);
			array->push(std::move(valueStr));
			++(*i);
		} else if (isdigit(str[*i])) {
			const int valueStart = *i;
			skipDigits(str, i);
			const int valueLength = *i - valueStart;
			const std::string valueStr = str.substr(valueStart, valueLength);
			array->push(std::stoi(valueStr));
		} else if (std::string_view{str}.substr(*i, 4) == "true") {
			const int valueStart = *i;
			*i += 4;
			const int valueLength = *i - valueStart;
			array->push(true);
		} else if (std::string_view{str}.substr(*i, 4) == "null") {
			*i += 4;
			array->pushNull();
		} else if (std::string_view{str}.substr(*i, 5) == "false") {
			const int valueStart = *i;
			*i += 5;
			const int valueLength = *i - valueStart;
			array->push(false);
		} else if (str[*i] == '{'){
			JsonObject newObject{}; 
			parseObject(str, i, &newObject);
			array->push(std::move(newObject));
		} else if (str[*i] == '['){
			JsonArray newArray{}; 
			parseArray(str, i, &newArray);
			array->push(std::move(newArray));
		}

	}

	void parseArray(const std::string &str, int *i, JsonArray* object) {
		if (str[*i] != '[') {
			throw "wrong json";
		}
		++(*i);
		skipWhiteSpaces(str, i);
		while (str[*i] != ']') {
			parseArrayField(str, i, object);
			skipWhiteSpaces(str, i);
			if (str[*i] == ',') {
				++(*i);
				skipWhiteSpaces(str, i);
				if (str[*i] == ']') {
					throw "wrong json: extra comma on last array field";
				}
			} else if (str[*i] != ']') {
				throw "wrong json: expected array end";
			}
		}
		++(*i);
	}
	
	void parseObject(const std::string &str, int *i, JsonObject* object) {
		if (str[*i] != '{') {
			throw "wrong json";
		}
		++(*i);
		skipWhiteSpaces(str, i);
		while (str[*i] != '}') {
			if (str[*i] != '"') {
				throw "wrong json: expect opening key quote";
			}
			parseObjectField(str, i, object);
			skipWhiteSpaces(str, i);
			if (str[*i] == ',') {
				++(*i);
				skipWhiteSpaces(str, i);
				if (str[*i] == '}') {
					throw "wrong json: extra comma on last object field";
				}
			} else if (str[*i] != '}') {
				const std::string sstr = str.substr(*i - 1);
				throw "wrong json: expected object end";
			}
		}
		++(*i);
		
	}

	std::variant<JsonObject, JsonArray> parse(const std::string& jsonStr) {
		int i = 0;
		skipWhiteSpaces(jsonStr, &i);
		if (jsonStr[i] == '{') {
			JsonObject object{};
			parseObject(jsonStr, &i, &object);
			return std::move(object);
		} else if (jsonStr[i] == '[') {
			JsonArray array{};
			parseArray(jsonStr, &i, &array);
			return std::move(array);
		}
		throw std::runtime_error("Bad json syntax");
	}
}
